"""
Music Assistant Client.

Simple wrapper for the websockets and rest Api's
provided by Music Assistant that allows for rapid development of apps
connected to Music Assistant.
"""

import asyncio
import functools
import logging
import uuid
from typing import Any, Awaitable, Callable, Dict, List, Optional, Union

from aiohttp import (
    ClientSession,
    ClientWebSocketResponse,
    TCPConnector,
    WSMsgType,
    client_exceptions,
)

APP_ID = "musicassistant-client"
LOGGER = logging.getLogger(APP_ID)

EVENT_CONNECTED = "connected"
EVENT_DISCONNECTED = "disconnected"
EVENT_PLAYER_ADDED = "player added"
EVENT_PLAYER_REMOVED = "player removed"
EVENT_PLAYER_CHANGED = "player changed"
EVENT_QUEUE_UPDATED = "queue updated"
EVENT_QUEUE_ITEMS_UPDATED = "queue items updated"
EVENT_SHUTDOWN = "application shutdown"
EVENT_PROVIDER_REGISTERED = "provider registered"
EVENT_PLAYER_CONTROL_REGISTERED = "player control registered"
EVENT_PLAYER_CONTROL_UNREGISTERED = "player control unregistered"
EVENT_PLAYER_CONTROL_UPDATED = "player control updated"
EVENT_SET_PLAYER_CONTROL_STATE = "set player control state"
EVENT_REGISTER_PLAYER_CONTROL = "register player control"
EVENT_UNREGISTER_PLAYER_CONTROL = "unregister player control"
EVENT_UPDATE_PLAYER_CONTROL = "update player control"

# pylint: disable=c-extension-no-member

SIZE_PARSE_JSON_EXECUTOR = 8192


class ConnectionClosed(Exception):
    """Exception raised when the connection is closed."""


class CannotConnect(Exception):
    """Exception raised when failed to connect the client."""

    def __init__(self, error: Exception) -> None:
        """Initialize a cannot connect error."""
        super().__init__(f"{error}", error)


class ConnectionFailed(Exception):
    """Exception raised when an established connection fails."""

    def __init__(self, error: Optional[Exception] = None) -> None:
        """Initialize a connection failed error."""
        if error is None:
            super().__init__("Connection failed.")
            return
        super().__init__(f"{error}", error)


class InvalidMessage(Exception):
    """Exception raised when an invalid message is received."""


class AuthenticationFailed(Exception):
    """Exception raised when authentication failed."""


class FailedCommand(Exception):
    """When a command has failed."""


class NotConnected(Exception):
    """Exception raised when trying to execute action while client is not connected."""


async def login(
    host: str, username: str, password: str, app_id: str = None, port: int = 8095
) -> dict:
    """
    Retrieve token to access a local MusicAssistant server.

        :param host: Hostname of Music Assistant Server.
        :param username: Username to authenticate.
        :param password: Password to authenticate.
        :param app_id: Some (friendly) identifier for your app.
                       A short-lived, single session token will be issued if appp_id is ommitted.
        :param port: The port of the server, defaults to 8095.
    """

    login_data = {"username": username, "password": password, "app_id": app_id}
    LOGGER.debug("Connecting to %s", host)
    async with ClientSession() as http_session:
        url = f"http://{host}:{port}/login"
        async with http_session.post(
            url, data=login_data, raise_for_status=True
        ) as resp:
            resp = await resp.json()
            LOGGER.debug(
                "Successfully logged in to Music Assistant server %s", resp["server_id"]
            )
            return resp


class MusicAssistant:
    """Connection to MusicAssistant server."""

    def __init__(
        self,
        host: str,
        token: str,
        port: int = 8095,
        loop: Optional[asyncio.AbstractEventLoop] = None,
        aiohttp_session: Optional[ClientSession] = None,
        auto_reconnect: bool = True,
    ) -> None:
        """
        Initialize the connection to a MusicAssistant server.

            :param hostname: Hostname of Music Assistant Server.
            :param token: (long lived) JWT token to use for authentication, may be retrieved with get_token().
            :param port: The port to use for this Music Assistant instance, default is 8095.
            :param loop: Optionally provide the event loop.
            :param client_session: Optionally provide a aiohttp ClientSession.
            :auto_reconnect: Automatically reconnect a lost session.
        """
        if isinstance(token, dict):
            self._token = token["token"]
        else:
            self._token = token
        self._ws_address = f"ws://{host}:{port}/ws"
        self._loop = loop
        if aiohttp_session:
            self.aiohttp_session_provided = True
            self.aiohttp_session = aiohttp_session
        else:
            self.aiohttp_session_provided = False
            self.aiohttp_session = None
        self.auto_reconnect = auto_reconnect
        self._client: Optional[ClientWebSocketResponse] = None
        self._result_futures: Dict[str, asyncio.Future] = {}
        self._shutdown_complete_event: Optional[asyncio.Event] = None
        self._event_listeners: list = []
        self._connected = False
        self._player_controls = {}
        self._server_info = {}
        self._pending_messages = []

    async def __aenter__(self):
        """Enter."""
        await self.connect()
        return self

    async def __aexit__(self, exc_type, exc_value, traceback):
        """Exit."""
        await self.disconnect()

    @property
    def connected(self):
        """Return bool if the connection with the server is alive."""
        return self._connected

    @property
    def server_info(self):
        """Return server info."""
        return self._server_info

    @property
    def address(self):
        """Return server address."""
        return self._server_info.get("address", self._address)

    @property
    def server_id(self):
        """Return server id."""
        return self._server_info.get("id")

    @property
    def server_version(self):
        """Return server version."""
        return self._server_info.get("version")

    @property
    def server_name(self):
        """Return server name."""
        return self._server_info.get("friendly_name")

    async def connect(self) -> None:
        """Connect to a local Music Assistant server."""
        if self._shutdown_complete_event is not None:
            # disconnect was requested
            return
        if not self._loop:
            self._loop = asyncio.get_event_loop()
        if self.aiohttp_session is None:
            self.aiohttp_session = ClientSession(
                loop=self._loop, connector=TCPConnector()
            )
        try:
            self._client = await self.aiohttp_session.ws_connect(
                self._ws_address,
                heartbeat=55,
            )
        except (
            client_exceptions.WSServerHandshakeError,
            client_exceptions.ClientError,
        ) as err:
            raise CannotConnect(err) from err

        # receive first message (=server info)
        json_msg = await self._receive_json_or_raise()
        self._server_info = json_msg["data"]

        # send auth message
        await self._send_auth()

        LOGGER.info(
            "Connected to Music Assistant %s (Version %s)",
            self.server_name,
            self.server_version,
        )

        # start task to keep listening for more messages
        asyncio.create_task(self._listen())

    async def disconnect(self) -> None:
        """Disconnect from the server."""

        LOGGER.debug("Disconnect requested")

        self._shutdown_complete_event = asyncio.Event()
        await self._client.close()
        await self._shutdown_complete_event.wait()

        if self.aiohttp_session and not self.aiohttp_session_provided:
            await self.aiohttp_session.close()
        self._shutdown_complete_event = None

    def register_event_callback(
        self,
        cb_func: Callable[..., Union[None, Awaitable]],
        event_filter: Union[None, str, List[str]] = None,
    ) -> Callable:
        """
        Add callback for events.

        Returns function to remove the listener.
            :param cb_func: callback function or coroutine
            :param event_filter: Optionally only listen for these events
        """
        listener = (cb_func, event_filter)
        self._event_listeners.append(listener)

        def remove_listener() -> None:
            self._event_listeners.remove(listener)

        return remove_listener

    async def register_player_control(
        self,
        control_type: int,
        control_id: str,
        provider_name: str,
        name: str,
        state: Any,
        cb_func: Callable[..., Union[None, Awaitable]],
    ) -> None:
        """
        Register a playercontrol with Music Assistant.

            :param control_type: Type of the PlayerControl. 0 for PowerControl, 1 for VolumeControl
            :param control_id: A unique id for this PlayerControl.
            :param provider_name: Your application name.
            :param name: A friendly name for this control.
            :param state: The current state for this control.
            :param cb_func: A callback that will be called when a new state needs to be set.
        """
        control = {
            "type": control_type,
            "control_id": control_id,
            "provider": provider_name,
            "name": name,
            "state": state,
        }
        self._player_controls[control_id] = (control, cb_func)
        await self.send_command(
            f"players/controls/{control_id}", control=control, method="POST"
        )

    async def update_player_control(self, control_id: str, new_state: Any) -> None:
        """
        Process updated state of an existing playercontrol.

            :param control_id: A unique id for this PlayerControl.
            :param new_state: The current/new state for this control.
        """
        if control_id not in self._player_controls:
            raise RuntimeError("Invalid player control")
        control = self._player_controls[control_id][0]
        control["state"] = new_state
        await self.send_command(
            f"players/controls/{control_id}", control=control, method="PUT"
        )

    async def get_library_artists(self) -> List[dict]:
        """Return all library artists on Music Assistant."""
        return await self.send_command("library/artists")

    async def get_library_albums(self) -> List[dict]:
        """Return all library albums on Music Assistant."""
        return await self.send_command("library/albums")

    async def get_library_tracks(self) -> List[dict]:
        """Return all library tracks on Music Assistant."""
        return await self.send_command("library/tracks")

    async def get_library_playlists(self) -> List[dict]:
        """Return all library playlists on Music Assistant."""
        return await self.send_command("library/playlists")

    async def get_library_radios(self) -> List[dict]:
        """Return all library radios on Music Assistant."""
        return await self.send_command("library/radios")

    async def get_item(self, item_id: str, provider_id: str, media_type: str):
        """Get single music item by id and media type."""
        return await self.send_command(f"items/{media_type}/{provider_id}/{item_id}")

    async def get_artist(self, artist_id: str, provider_id: str) -> dict:
        """Return full artist object for specified artist/provider id.."""
        return await self.send_command(f"artists/{provider_id}/{artist_id}")

    async def get_album(self, album_id: str, provider_id: str) -> dict:
        """Return full album object for specified album/provider id.."""
        return await self.send_command(f"albums/{provider_id}/{album_id}")

    async def get_track(self, track_id: str, provider_id: str) -> dict:
        """Return full track object for specified track/provider id.."""
        return await self.send_command(f"tracks/{provider_id}/{track_id}")

    async def get_playlist(self, playlist_id: str, provider_id: str) -> dict:
        """Return full playlist object for specified playlist/provider id.."""
        return await self.send_command(f"playlists/{provider_id}/{playlist_id}")

    async def get_radio(self, radio_id: str, provider_id: str) -> dict:
        """Return full radio object for specified radio/provider id.."""
        return await self.send_command(f"radios/{provider_id}/{radio_id}")

    async def get_image_thumb(
        self,
        media_item: Optional[dict] = None,
        url: Optional[str] = None,
        size: int = 500,
    ) -> str:
        """Return base64 thumbnail image (data) for the given media item OR url."""
        if not url:
            url = await self.get_media_item_image_url(media_item)
        return await self.send_command("images/thumb", url=url, size=size)

    async def get_media_item_image_url(self, media_item: dict):
        """Get full/original image url for given media_item by providing the media item."""
        if not media_item:
            return None
        if media_item.get("metadata") and media_item["metadata"].get("image"):
            return media_item["metadata"]["image"]
        if media_item.get("album") and media_item["album"].get("metadata"):
            if media_item["album"]["metadata"].get("image"):
                return media_item["album"]["metadata"]["image"]
        if media_item.get("artist") and media_item["artist"].get("metadata"):
            if media_item["artist"]["metadata"].get("image"):
                return media_item["artist"]["metadata"]["image"]
        return await self.send_command(
            f'images/{media_item["media_type"]}/{media_item["provider"]}/{media_item["item_id"]}'
        )

    async def get_artist_toptracks(
        self, artist_id: str, provider_id: str
    ) -> List[dict]:
        """Return top tracks for specified artist/provider id."""
        return await self.send_command(f"artists/{provider_id}/{artist_id}/tracks")

    async def get_artist_albums(self, artist_id: str, provider_id: str) -> List[dict]:
        """Return albums for specified artist/provider id."""
        return await self.send_command(f"artists/{provider_id}/{artist_id}/albums")

    async def get_playlist_tracks(
        self, playlist_id: str, provider_id: str
    ) -> List[dict]:
        """Return the playlist's tracks for specified playlist/provider id."""
        return await self.send_command(f"playlists/{provider_id}/{playlist_id}/tracks")

    async def get_album_tracks(self, album_id: str, provider_id: str) -> List[dict]:
        """Return the album's tracks for specified album/provider id."""
        return await self.send_command(f"albums/{provider_id}/{album_id}/tracks")

    async def search(
        self, search_query: str, media_types: List[str] = None, limit=10
    ) -> dict:
        """
        Perform global search for media items on all providers.

            :param search_query: Search query.
            :param media_types: A list of media_types to include.
            :param limit: number of items to return in the search (per type). All if ommitted.
        """
        if media_types is None:
            media_types = ["artists", "albums", "tracks", "playlists", "radios"]
        return await self.send_command(
            "search", search_query=search_query, media_types=media_types, limit=limit
        )

    async def get_players(self) -> List[dict]:
        """Return all players on Music Assistant."""
        return await self.send_command("players")

    async def get_player(self, player_id: str) -> dict:
        """Return player details for the given player."""
        return await self.send_command(f"players/{player_id}")

    async def get_player_queues(self) -> List[dict]:
        """Return all player queues on Music Assistant."""
        return await self.send_command("queues")

    async def get_player_queue(self, queue_id: str) -> dict:
        """Return queue details for the given playerqueue."""
        return await self.send_command(f"queues/{queue_id}")

    async def get_player_queue_items(self, queue_id: str) -> List[dict]:
        """Return all queue items for the given player."""
        return await self.send_command(f"queues/{queue_id}/items")

    async def player_queue_set_shuffle(
        self, queue_id: str, enable_shuffle: bool = False
    ) -> None:
        """Send enable/disable shuffle command to given playerqueue."""
        await self.send_command(
            f"queues/{queue_id}", enable_shuffle=enable_shuffle, method="PUT"
        )

    async def player_queue_set_repeat(
        self, queue_id: str, enable_repeat: bool = False
    ) -> None:
        """Send enable/disable repeat command to given playerqueue."""
        await self.send_command(
            f"queues/{queue_id}", enable_repeat=enable_repeat, method="PUT"
        )

    async def player_queue_clear(self, queue_id: str) -> None:
        """Clear all items in the player's queue."""
        await self.send_command(f"queues/{queue_id}/items", method="DELETE")

    async def player_queue_move_item(
        self, queue_id: str, queue_item_id: str, pos_shift: int = 1
    ) -> None:
        """
        Move queue item x up/down the queue.

        param pos_shift: move item x positions down if positive value
                         move item x positions up if negative value
                         move item to top of queue as next item if 0
        """
        await self.send_command(
            f"queues/{queue_id}/cmd/move",
            queue_item_id=queue_item_id,
            pos_shift=pos_shift,
            method="PUT",
        )

    async def play_index(self, queue_id: str, index: Union[int, str]) -> None:
        """Play item at index (or item_id) X in queue."""
        await self.send_command(
            f"queues/{queue_id}/cmd/play_index", index=index, method="PUT"
        )

    async def player_command(self, player_id, cmd: str, **kwargs) -> None:
        """Execute command on given player."""
        await self.send_command(
            f"players/{player_id}/cmd/{cmd}", method="PUT", **kwargs
        )

    async def play_media(
        self,
        player_id: str,
        media_items: Union[dict, List[dict]],
        queue_opt: str = "play",
    ) -> None:
        """
        Play media item(s) on the given player.

            :param player_id: player_id of the player to handle the command.
            :param media_item: media item(s) that should be played (single item or list of items)
            :param queue_opt:
                play -> Insert new items in queue and start playing at inserted position
                replace -> Replace queue contents with these items
                next -> Play item(s) after current playing item
                add -> Append new items at end of the queue
        """
        await self.send_command(
            f"players/{player_id}/play_media",
            items=media_items,
            queue_opt=queue_opt,
            method="PUT",
        )

    async def play_uri(
        self,
        player_id: str,
        uri: str,
        queue_opt: str = "play",
    ) -> None:
        """
        Play the specified uri/url on the given player.

        Either specify a valid (Mass supported) MediaItem URI or an URL to some audio.

            :param player_id: player_id of the player to handle the command.
            :param uri: Url/Uri that can be played by a player.
        """
        await self.send_command(
            f"players/{player_id}/play_uri",
            uri=uri,
            queue_opt=queue_opt,
            method="PUT",
        )

    async def play_alert(
        self,
        player_id: str,
        url: str,
        volume: Optional[int] = None,
        force: bool = True,
        announce: bool = False,
    ) -> None:
        """
        Play alert (e.g. tts message) on selected player.

        Will pause the current playing queue and resume after the alert is played.

            :param player_id: player_id of the player to handle the command.
            :param url: Url to the sound effect/tts message that should be played.
            :param volume: Force volume of player to this level during the alert.
            :param force: Play alert even if player is currently powered off.
            :param announce: Announce the alert by prepending an alert sound.
        """
        await self.send_command(
            f"players/{player_id}/play_alert",
            url=url,
            volume=volume,
            force=force,
            announce=announce,
            method="PUT",
        )

    async def send_command(self, command: str, **kwargs) -> dict:
        """Send a command and get a response."""
        future: "asyncio.Future[dict]" = self._loop.create_future()
        message_id = uuid.uuid4().hex
        self._result_futures[message_id] = future
        kwargs["path"] = command
        if "method" not in kwargs:
            kwargs["method"] = "GET"
        await self._send_json_message(
            {"id": message_id, "type": "command", "data": kwargs}
        )
        try:
            return await future
        finally:
            self._result_futures.pop(message_id, None)

    async def _listen(self) -> None:
        """Start listening to the websocket."""
        if not self._connected:
            raise NotConnected("Not connected when start listening")

        assert self._client
        try:
            while not self._client.closed:
                data = await self._receive_json_or_raise()
                await self._handle_incoming_message(data)
        except ConnectionClosed:
            pass
        finally:
            self._connected = False
            LOGGER.debug("Disconnected from websocket. Cleaning up")

            for future in self._result_futures.values():
                future.cancel()
            self._result_futures.clear()

            if not self._client.closed:
                await self._client.close()

            if self._shutdown_complete_event:
                # shutdown was requested (and is awaited for)
                self._shutdown_complete_event.set()
            else:
                # unexpected loss of connection ?
                await self._signal_event(EVENT_DISCONNECTED)
                if self.auto_reconnect:
                    # schedule reconnect
                    asyncio.create_task(self._auto_reconnect())

    async def _send_auth(
        self,
    ) -> dict:
        """Send auth request."""
        message_id = uuid.uuid4().hex
        await self._client.send_json(
            {"id": message_id, "type": "auth", "data": self._token}
        )
        json_msg = await self._receive_json_or_raise()
        if json_msg["type"] == "error":
            raise AuthenticationFailed(json_msg["data"])
        if json_msg["id"] != message_id:
            raise AuthenticationFailed("Unexpected response")
        # Connected !
        self._connected = True
        await self._signal_event(EVENT_CONNECTED)
        # send any queued messages
        for message in self._pending_messages:
            await self._client.send_json(message)
        self._pending_messages = []

    async def _receive_json_or_raise(self) -> dict:
        """Receive json or raise."""
        assert self._client
        msg = await self._client.receive()

        if msg.type in (WSMsgType.CLOSE, WSMsgType.CLOSED, WSMsgType.CLOSING):
            raise ConnectionClosed("Connection was closed.")

        if msg.type == WSMsgType.ERROR:
            raise ConnectionFailed()

        if msg.type != WSMsgType.TEXT:
            raise InvalidMessage(f"Received non-Text message: {msg.type}")

        try:
            if len(msg.data) > SIZE_PARSE_JSON_EXECUTOR:
                data: dict = await self._loop.run_in_executor(None, msg.json)
            else:
                data = msg.json()
        except ValueError as err:
            raise InvalidMessage("Received invalid JSON.") from err

        return data

    async def _handle_incoming_message(self, msg: dict) -> None:
        """Handle incoming message."""

        if msg["type"] == "event":
            event = msg["data"]["event"]
            event_data = msg["data"].get("event_data")
            await self._signal_event(event, event_data)
            return

        future = self._result_futures.get(msg["id"])
        if msg["type"] in ["result", "error"] and future is None:
            # no listener for this result
            return

        if msg["type"] == "error":
            err = FailedCommand(msg["data"])
            future.set_exception(err)
            return

        if msg["type"] == "result":
            future.set_result(msg["data"])
            return

        if msg["type"] == "command":
            # command from server to this client
            command = msg["data"]["path"]
            params = {
                x: msg["data"][x] for x in msg["data"] if x not in ["path", "method"]
            }
            if "players/controls/" in command and "/state" in command:
                # player control state request
                control_id = command.split("/")[2]
                control = self._player_controls.get(control_id)
                if control:
                    await control[1](params)
                return
            # simply broadcast the command as event
            await self._signal_event(command, params)
            return

        # Can't handle
        LOGGER.debug(
            "Received message with unknown type '%s': %s",
            msg["type"],
            msg,
        )

    async def _send_json_message(self, message: Dict[str, Any]) -> None:
        """Send a message.

        Raises NotConnected if client not connected.
        If auto reconnect is enabled, messages will be queued and sent when client is (re)connected.
        """

        assert self._client
        assert "id" in message

        if not self._connected:
            # trying to send message when not connected
            if self.auto_reconnect:
                # put message in queue, to be sent when (re)connected
                self._pending_messages.append(message)
                return
            raise NotConnected

        await self._client.send_json(message)

    async def _signal_event(self, event: str, event_details: Any = None) -> None:
        """Signal event to registered callbacks."""
        for cb_func, event_filter in self._event_listeners:
            if not (event_filter is None or event in event_filter):
                continue
            # call callback
            check_target = cb_func
            while isinstance(check_target, functools.partial):
                check_target = check_target.func
                check_target.args = (event, event_details)
            if asyncio.iscoroutine(check_target):
                self._loop.create_task(check_target)
            elif asyncio.iscoroutinefunction(check_target):
                self._loop.create_task(check_target(event, event_details))
            else:
                self._loop.run_in_executor(None, cb_func, event, event_details)

    async def _auto_reconnect(self):
        """Reconnect to the server on lost connection."""
        while not self._shutdown_complete_event:
            try:
                LOGGER.debug("Trying to reconnect to the server...")
                await self.connect()
                break
            except CannotConnect:
                await asyncio.sleep(30)
            # fails on all other errors (e.g. Auth error)
